//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Interface/AutocorrelationModels.h
//! @brief     Define AutocorrelationModel classes.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2024
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_INTERFACE_AUTOCORRELATIONMODELS_H
#define BORNAGAIN_SAMPLE_INTERFACE_AUTOCORRELATIONMODELS_H

#include "Base/Type/ICloneable.h"
#include "Param/Node/INode.h"
#include <heinz/Vectors3D.h>

//! Base class for autocorrelation function models.
class AutocorrelationModel : public ICloneable, public INode {
public:
    AutocorrelationModel* clone() const override = 0;
    virtual double sigma() const = 0;
    virtual double spectralFunction(const R3& k) const = 0;
    virtual double corrFunction(const R3& k) const = 0;

#ifndef SWIG
    //! Used to create the Python constructor of this class
    virtual std::string pythonArguments() const = 0;
#endif
};

//! The model is described e.g. in the article "Roughness spectrum and surface width of self-affine
//! fractal surfaces via the K-correlation model" by G. Palasantzas, Physical Review B 48,
//! 14472-14478 (1993)
class K_CorrelationModel : public AutocorrelationModel {
public:
    K_CorrelationModel(double sigma = 0, double hurst = 0, double lateralCorrLength = 0);
    K_CorrelationModel* clone() const override;
    std::string className() const override { return "K_CorrelationModel"; }
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Sigma", "nm"}, {"Hurst", ""}, {"CorrLength", "nm"}};
    }
    std::string validate() const override;

#ifndef SWIG
    std::string pythonArguments() const override;
#endif

    //! Returns power spectral density of the surface roughness
    double spectralFunction(const R3& k) const override;

    double corrFunction(const R3& k) const override;

    //! Sets rms of roughness
    void setSigma(double sigma)
    {
        m_sigma = sigma;
    }
    //! Returns rms of roughness
    double sigma() const override
    {
        return m_sigma;
    }

    //! Sets hurst parameter. It describes how jagged the surface is.
    void setHurstParameter(double hurstParameter)
    {
        m_hurst_parameter = hurstParameter;
    }
    //! Returns hurst parameter
    double hurst() const
    {
        return m_hurst_parameter;
    }

    //! Sets lateral correlation length
    void setLatteralCorrLength(double lateralCorrLen)
    {
        m_lateral_corr_length = lateralCorrLen;
    }
    //! Returns lateral correlation length
    double lateralCorrLength() const
    {
        return m_lateral_corr_length;
    }

private:
    double m_sigma;           //!< rms of roughness
    double m_hurst_parameter; //!< Hurst parameter which describes how jagged the interface, 0<H<=1
    double m_lateral_corr_length; //!< lateral correlation length of the roughness
};

#endif // BORNAGAIN_SAMPLE_INTERFACE_AUTOCORRELATIONMODELS_H
