//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Interface/CrosscorrelationModels.cpp
//! @brief     Implement CrossCorrModel classes.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2024
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Interface/CrosscorrelationModels.h"
#include "Base/Py/PyFmt.h"
#include "Base/Util/Assert.h"
#include <cmath>

CommonDepthCrosscorrelation::CommonDepthCrosscorrelation(double cross_corr_depth)
    : m_crosscorrelation_depth(cross_corr_depth)
{
    validateOrThrow();
}

CommonDepthCrosscorrelation* CommonDepthCrosscorrelation::clone() const
{
    return new CommonDepthCrosscorrelation(m_crosscorrelation_depth);
}

std::string CommonDepthCrosscorrelation::validate() const
{
    std::vector<std::string> errs;
    requestGe0(errs, m_crosscorrelation_depth, "crossCorrDepth");
    if (!errs.empty())
        return jointError(errs);
    m_validated = true;
    return "";
}

std::string CommonDepthCrosscorrelation::pythonArguments() const
{
    return Py::Fmt::printArguments({{m_crosscorrelation_depth, parDefs()[0].unit}});
}

double CommonDepthCrosscorrelation::replicationFactor(const R3&, double thickness) const
{
    ASSERT(thickness >= 0);

    if (m_crosscorrelation_depth == 0)
        return 0;

    return std::exp(-thickness / m_crosscorrelation_depth);
}
